@@ -1,3 +1,4 @@
+// -*-  indent-tabs-mode:nil; tab-width:2;  -*-
 //! Utilities for reading and writing Java properties files
 //!
 //! The specification is taken from https://docs.oracle.com/javase/7/docs/api/java/util/Properties.html.
@@ -526,10 +527,20 @@ fn unicode_escape(_encoder: &mut RawEncoder, input: &str, output: &mut ByteWrite
 
 static UNICODE_ESCAPE: EncoderTrap = EncoderTrap::Call(unicode_escape);
 
+#[derive(PartialEq,Eq,PartialOrd,Ord,Debug,Copy,Clone)]
+pub enum LineEnding {
+  CR,
+  LF,
+  CRLF,
+}
+
 /// Writes to a properties file.
 pub struct PropertiesWriter<W: Write> {
   writer: W,
   lines_written: usize,
+  comment_prefix: Vec<u8>,
+  kv_separator: Vec<u8>,
+  line_ending: LineEnding,
 }
 
 impl<W: Write> PropertiesWriter<W> {
@@ -538,19 +549,31 @@ impl<W: Write> PropertiesWriter<W> {
     PropertiesWriter {
       writer: writer,
       lines_written: 0,
+      comment_prefix: vec!['#' as u8, ' ' as u8],
+      kv_separator: vec!['=' as u8],
+      line_ending: LineEnding::LF,
     }
   }
 
+  fn write_eol(&mut self) -> Result<(), PropertiesError> {
+    try!(match self.line_ending {
+      LineEnding::CR => self.writer.write_all(&['\r' as u8]),
+      LineEnding::LF => self.writer.write_all(&['\n' as u8]),
+      LineEnding::CRLF => self.writer.write_all(&['\r' as u8, '\n' as u8]),
+    });
+    Ok(())
+  }
+
   /// Writes a comment to the file.
   pub fn write_comment(&mut self, comment: &str) -> Result<(), PropertiesError> {
     self.lines_written += 1;
-    try!(self.writer.write_all(&['#' as u8, ' ' as u8]));
+    try!(self.writer.write_all(&self.comment_prefix));
     let data = ISO_8859_1.encode(comment, UNICODE_ESCAPE);
     match data {
       Ok(d) => try!(self.writer.write_all(&d)),
       Err(_) => return Err(PropertiesError::new("Encoding error", None, Some(self.lines_written))),
     };
-    try!(self.writer.write_all(&['\n' as u8]));
+    try!(self.write_eol());
     Ok(())
   }
 
@@ -581,9 +604,9 @@ impl<W: Write> PropertiesWriter<W> {
   /// Writes a key/value pair to the file.
   pub fn write(&mut self, key: &str, value: &str) -> Result<(), PropertiesError> {
     try!(self.write_escaped(key));
-    try!(self.writer.write_all(&['=' as u8]));
+    try!(self.writer.write_all(&self.kv_separator));
     try!(self.write_escaped(value));
-    try!(self.writer.write_all(&['\n' as u8]));
+    try!(self.write_eol());
     Ok(())
   }
 
@@ -592,6 +615,45 @@ impl<W: Write> PropertiesWriter<W> {
     try!(self.writer.flush());
     Ok(())
   }
+
+  /// Sets the comment prefix.
+  ///
+  /// The prefix must contain a '#' or a '!', may only contain spaces, tabs, or form feeds before the comment character,
+  /// and may not contain any carriage returns or line feeds ('\r' or '\n').
+  pub fn set_comment_prefix(&mut self, prefix: &str) -> Result<(), PropertiesError> {
+    let re = Regex::new(r"^[ \t\x0c]*[#!][^\r\n]*$").unwrap();
+    if !re.is_match(prefix) {
+      return Err(PropertiesError::new(&format!("Bad comment prefix: {:?}", prefix), None, None));
+    }
+    let data = ISO_8859_1.encode(prefix, UNICODE_ESCAPE);
+    match data {
+      Ok(bytes) => self.comment_prefix = bytes,
+      Err(_) => return Err(PropertiesError::new("Encoding error", None, None)),
+    };
+    Ok(())
+  }
+
+  /// Sets the key/value separator.
+  ///
+  /// The separator may be non-empty whitespace, or a colon with optional whitespace on either side,
+  /// or an equals sign with optional whitespace on either side.  (Whitespace here means ' ', '\t', or '\f'.)
+  pub fn set_kv_separator(&mut self, separator: &str) -> Result<(), PropertiesError> {
+    let re = Regex::new(r"^([ \t\x0c]*[:=][ \t\x0c]*|[ \t\x0c]+)$").unwrap();
+    if !re.is_match(separator) {
+      return Err(PropertiesError::new(&format!("Bad key/value separator: {:?}", separator), None, None));
+    }
+    let data = ISO_8859_1.encode(separator, UNICODE_ESCAPE);
+    match data {
+      Ok(bytes) => self.kv_separator = bytes,
+      Err(_) => return Err(PropertiesError::new("Encoding error", None, None)),
+    };
+    Ok(())
+  }
+
+  /// Sets the line ending.
+  pub fn set_line_ending(&mut self, line_ending: LineEnding) {
+    self.line_ending = line_ending;
+  }
 }
 
 /////////////////////
@@ -607,6 +669,7 @@ mod tests {
   use super::CR;
   use super::LF;
   use super::Line;
+  use super::LineEnding;
   use super::LogicalLine;
   use super::LogicalLines;
   use super::NaturalLine;
@@ -871,6 +934,135 @@ mod tests {
     }
   }
 
+  #[test]
+  fn properties_writer_good_comment_prefix() {
+    let prefixes = ["#", "!", " #", " !", "#x", "!x", "\x0c#"];
+    let mut buf = Vec::new();
+    for prefix in prefixes.iter() {
+      let mut writer = PropertiesWriter::new(&mut buf);
+      writer.set_comment_prefix(prefix).unwrap();
+    }
+  }
+
+  #[test]
+  fn properties_writer_bad_comment_prefix() {
+    let prefixes = ["", " ", "x", "\n#", "#\n", "#\r"];
+    let mut buf = Vec::new();
+    for prefix in prefixes.iter() {
+      let mut writer = PropertiesWriter::new(&mut buf);
+      match writer.set_comment_prefix(prefix) {
+        Ok(_) => panic!("Unexpectedly succeded with prefix {:?}", prefix),
+        Err(_) => (),
+      }
+    }
+  }
+
+  #[test]
+  fn properties_writer_custom_comment_prefix() {
+    let data = [
+      ("", " !\n"),
+      ("a", " !a\n"),
+      (" :=", " ! :=\n"),
+      ("\u{1F41E}", " !\\u1f41e\n"),
+    ];
+    for &(comment, expected) in data.iter() {
+      let mut buf = Vec::new();
+      {
+        let mut writer = PropertiesWriter::new(&mut buf);
+        writer.set_comment_prefix(" !").unwrap();
+        writer.write_comment(comment).unwrap();
+      }
+      match ISO_8859_1.decode(&buf, DecoderTrap::Strict) {
+        Ok(actual) => {
+          if expected != actual {
+            panic!("Failure while processing {:?}.  Expected {:?}, but was {:?}", comment, expected, actual);
+          }
+        },
+        Err(_) => panic!("Error decoding test output"),
+      }
+    }
+  }
+
+  #[test]
+  fn properties_writer_good_kv_separator() {
+    let separators = [":", "=", " ", " :", ": ", " =", "= ", "\x0c", "\t"];
+    let mut buf = Vec::new();
+    for separator in separators.iter() {
+      let mut writer = PropertiesWriter::new(&mut buf);
+      writer.set_kv_separator(separator).unwrap();
+    }
+  }
+
+  #[test]
+  fn properties_writer_bad_kv_separator() {
+    let separators = ["", "x", ":=", "=:", "\n", "\r"];
+    let mut buf = Vec::new();
+    for separator in separators.iter() {
+      let mut writer = PropertiesWriter::new(&mut buf);
+      match writer.set_kv_separator(separator) {
+        Ok(_) => panic!("Unexpectedly succeded with separator {:?}", separator),
+        Err(_) => (),
+      }
+    }
+  }
+
+  #[test]
+  fn properties_writer_custom_kv_separator() {
+    let data = [
+      (":", "x:y\n"),
+      ("=", "x=y\n"),
+      (" ", "x y\n"),
+      (" :", "x :y\n"),
+      (": ", "x: y\n"),
+      (" =", "x =y\n"),
+      ("= ", "x= y\n"),
+      ("\x0c", "x\x0cy\n"),
+      ("\t", "x\ty\n")
+    ];
+    for &(separator, expected) in data.iter() {
+      let mut buf = Vec::new();
+      {
+        let mut writer = PropertiesWriter::new(&mut buf);
+        writer.set_kv_separator(separator).unwrap();
+        writer.write("x", "y").unwrap();
+      }
+      match ISO_8859_1.decode(&buf, DecoderTrap::Strict) {
+        Ok(actual) => {
+          if expected != actual {
+            panic!("Failure while processing {:?}.  Expected {:?}, but was {:?}", separator, expected, actual);
+          }
+        },
+        Err(_) => panic!("Error decoding test output"),
+      }
+    }
+  }
+
+  #[test]
+  fn properties_writer_custom_line_ending() {
+    let data = [
+      (LineEnding::CR, "# foo\rx=y\r"),
+      (LineEnding::LF, "# foo\nx=y\n"),
+      (LineEnding::CRLF, "# foo\r\nx=y\r\n"),
+    ];
+    for &(line_ending, expected) in data.iter() {
+      let mut buf = Vec::new();
+      {
+        let mut writer = PropertiesWriter::new(&mut buf);
+        writer.set_line_ending(line_ending);
+        writer.write_comment("foo").unwrap();
+        writer.write("x", "y").unwrap();
+      }
+      match ISO_8859_1.decode(&buf, DecoderTrap::Strict) {
+        Ok(actual) => {
+          if expected != actual {
+            panic!("Failure while processing {:?}.  Expected {:?}, but was {:?}", line_ending, expected, actual);
+          }
+        },
+        Err(_) => panic!("Error decoding test output"),
+      }
+    }
+  }
+
   struct ErrorReader;
 
   impl Read for ErrorReader {
